from __future__ import annotations

import dataclasses as dc
import logging
import time
from abc import ABCMeta, abstractmethod
from typing import Any, Iterable, Mapping, MutableMapping, Optional

from .errors import MetabaseStateError
from .format import Filter, NullValue, safe_name
from .manifest import DEFAULT_SCHEMA, Column, Group, Manifest, Model
from .metabase import Metabase

_logger = logging.getLogger(__name__)


class ModelsMixin(metaclass=ABCMeta):
    """Abstraction for exporting models."""

    __SYNC_PERIOD = 5

    DEFAULT_MODELS_SYNC_TIMEOUT = 30

    @property
    @abstractmethod
    def manifest(self) -> Manifest:
        pass

    @property
    @abstractmethod
    def metabase(self) -> Metabase:
        pass

    def export_models(
        self,
        metabase_database: str,
        database_filter: Optional[Filter] = None,
        schema_filter: Optional[Filter] = None,
        model_filter: Optional[Filter] = None,
        skip_sources: bool = False,
        sync_timeout: int = DEFAULT_MODELS_SYNC_TIMEOUT,
        append_tags: bool = False,
        docs_url: Optional[str] = None,
        order_fields: bool = False,
    ):
        """Exports dbt models to Metabase database schema.

        Args:
            metabase_database (str): Target database in Metabase.
            database_filter (Optional[Filter], optional): Filter dbt manifest by database. Defaults to None.
            schema_filter (Optional[Filter], optional): Filter dbt manifest by schema. Defaults to None.
            model_filter (Optional[Filter], optional): Filter dbt manifest by model. Defaults to None.
            skip_sources (bool, optional): Exclude dbt sources from export. Defaults to False.
            sync_timeout (int, optional): Number of seconds to wait until Metabase schema matches the dbt project. To skip synchronization, set timeout to 0. Defaults to 30.
            append_tags (bool, optional): Append dbt tags to table descriptions. Defaults to False.
            docs_url (Optional[str], optional): URL for dbt docs hosting, to append model links to table descriptions. Defaults to None.
            order_fields (bool, optional): Preserve column order in dbt project.
        """

        ctx = self.__Context()
        success = True

        database = self.metabase.find_database(name=metabase_database)
        if not database:
            raise MetabaseStateError(f"Database not found: {metabase_database}")

        models = self.__filtered_models(
            models=self.manifest.read_models(),
            database_filter=database_filter,
            schema_filter=schema_filter,
            model_filter=model_filter,
            skip_sources=skip_sources,
        )

        self.metabase.sync_database_schema(database["id"])

        deadline = int(time.time()) + sync_timeout
        synced = False
        while not synced:
            time.sleep(self.__SYNC_PERIOD)

            tables = self.__get_tables(database["id"])

            synced = True
            for model in models:
                schema_name = model.schema.upper()
                model_name = model.alias.upper()
                table_key = f"{schema_name}.{model_name}"

                table = tables.get(table_key)
                if not table:
                    _logger.warning(
                        "Table '%s' not in schema '%s'", table_key, schema_name
                    )
                    synced = False
                    continue

                for column in model.columns:
                    column_name = column.name.upper()

                    field = table.get("fields", {}).get(column_name)
                    if not field:
                        _logger.warning(
                            "Field '%s' not in table '%s'", column_name, table_key
                        )
                        synced = False
                        continue

            ctx.tables = tables

            if int(time.time()) > deadline:
                break

        if not synced and sync_timeout:
            raise MetabaseStateError("Unable to sync models with Metabase")

        for model in models:
            success &= self.__export_model(
                ctx=ctx,
                model=model,
                append_tags=append_tags,
                docs_url=docs_url,
                order_fields=order_fields,
            )

        for update in ctx.updates.values():
            if update["kind"] == "table":
                self.metabase.update_table(
                    uid=update["id"],
                    body=update["body"],
                )
            elif update["kind"] == "table_field_order":
                self.metabase.update_table_field_order(
                    uid=update["id"],
                    body=list(update["body"]["values"]),
                )
            elif update["kind"] == "field":
                self.metabase.update_field(
                    uid=update["id"],
                    body=update["body"],
                )

            _logger.info(
                "%s '%s' updated successfully: %s",
                " ".join(update["kind"].split("_")).capitalize(),
                update["label"],
                ", ".join(update.get("body", {})),
            )

        if not success:
            raise MetabaseStateError("Non-critical errors encountered, see above")

    def __export_model(
        self,
        ctx: __Context,
        model: Model,
        append_tags: bool,
        docs_url: Optional[str],
        order_fields: bool,
    ) -> bool:
        """Exports one dbt model to Metabase database schema."""

        success = True

        schema_name = model.schema.upper()
        model_name = model.alias.upper()
        table_key = f"{schema_name}.{model_name}"

        api_table = ctx.tables.get(table_key)
        if not api_table:
            _logger.error("Table '%s' does not exist", table_key)
            return False

        # Empty strings not accepted by Metabase
        model_display_name = model.display_name or None
        model_description = model.format_description(append_tags, docs_url) or None
        model_points_of_interest = model.points_of_interest or None
        model_caveats = model.caveats or None
        model_visibility = model.visibility_type or None

        body_table = {}

        # Update if specified, otherwise reset one that had been set
        api_display_name = api_table.get("display_name")
        if api_display_name != model_display_name and (
            model_display_name
            or safe_name(api_display_name) != safe_name(api_table.get("name"))
        ):
            body_table["display_name"] = model_display_name

        if api_table.get("description") != model_description:
            body_table["description"] = model_description
        if api_table.get("points_of_interest") != model_points_of_interest:
            body_table["points_of_interest"] = model_points_of_interest
        if api_table.get("caveats") != model_caveats:
            body_table["caveats"] = model_caveats
        if api_table.get("visibility_type") != model_visibility:
            body_table["visibility_type"] = model_visibility

        if body_table:
            ctx.update(entity=api_table, change=body_table, label=table_key)
            _logger.info(
                "Table '%s' will be updated: %s", table_key, ", ".join(body_table)
            )
        else:
            _logger.info("Table '%s' is up to date", table_key)

        for column in model.columns:
            success &= self.__export_column(
                ctx,
                schema_name=schema_name,
                model_name=model_name,
                column=column,
            )

        if order_fields:
            success &= self.__export_model_column_order(
                ctx=ctx,
                model=model,
                api_table=api_table,
                table_key=table_key,
            )

        return success

    def __export_model_column_order(
        self,
        ctx: __Context,
        model: Model,
        api_table: Mapping,
        table_key: str,
    ) -> bool:
        """Exports model column order to Metabase field order."""

        api_ord = {}
        for field in api_table.get("fields", {}).values():
            api_ord[field["id"]] = field["name"]

        dbt_ord = {}
        for column in model.columns:
            field_id = ctx.get_field(table_key, column.name.upper()).get("id")
            if field_id:
                dbt_ord[field_id] = column.name

        if (
            list(api_ord.keys()) != list(dbt_ord.keys())
            or api_table.get("field_order") != "custom"
        ):
            api_keys = set(api_ord.keys())
            dbt_keys = set(dbt_ord.keys())

            if api_keys == dbt_keys:
                ctx.update(
                    entity={
                        "kind": "table_field_order",
                        "id": api_table["id"],
                    },
                    change={"values": dbt_ord.keys()},
                    label=table_key,
                )
                _logger.info("Table field order '%s' will be updated", table_key)
            else:
                details = []

                api_only = [api_ord[x] for x in api_keys - dbt_keys]
                if api_only:
                    details.append(f"missing in dbt {api_only}")

                dbt_only = [dbt_ord[x] for x in dbt_keys - api_keys]
                if dbt_only:
                    details.append(f"missing in Metabase {dbt_only}")

                _logger.error(
                    "Table field order '%s' mismatch: %s",
                    table_key,
                    ", ".join(details),
                )
                return False

        return True

    def __export_column(
        self,
        ctx: __Context,
        schema_name: str,
        model_name: str,
        column: Column,
    ) -> bool:
        """Exports one dbt column to Metabase database schema."""

        success = True

        table_key = f"{schema_name}.{model_name}"
        column_name = column.name.upper()
        column_label = f"{schema_name}.{model_name}.{column_name}"

        api_field = ctx.get_field(table_key, column_name)
        if not api_field:
            _logger.error("Field '%s.%s' does not exist", table_key, column_name)
            return False

        if "special_type" in api_field:
            semantic_type_key = "special_type"
        else:
            semantic_type_key = "semantic_type"

        fk_target_field_id = None
        if column.semantic_type == "type/FK":
            # Target table could be aliased if we parse_ref() on a source, so we caught aliases during model parsing
            # This way we can unpack any alias mapped to fk_target_table when using yml project reader
            fk_target_table_name = (
                column.fk_target_table.upper() if column.fk_target_table else None
            )
            fk_target_field_name = (
                column.fk_target_field.upper() if column.fk_target_field else None
            )
            fk_target_field_label = f"{fk_target_table_name}.{fk_target_field_name}"

            if fk_target_table_name and fk_target_field_name:
                fk_target_field = ctx.get_field(
                    table_key=fk_target_table_name,
                    field_key=fk_target_field_name,
                )
                if fk_target_field:
                    fk_target_field_id = fk_target_field.get("id")
                    if fk_target_field.get(semantic_type_key) != "type/PK":
                        _logger.info(
                            "Field '%s' will be updated as primary key for foreign key '%s'",
                            fk_target_field_label,
                            column_label,
                        )
                        ctx.update(
                            entity=fk_target_field,
                            change={semantic_type_key: "type/PK"},
                            label=fk_target_field_label,
                        )
                else:
                    _logger.error(
                        "Field '%s' referenced as foreign key '%s' not found",
                        fk_target_field_label,
                        column_label,
                    )
                    success = False
            else:
                _logger.info(
                    "Field '%s' referenced as foreign key '%s' is invalid",
                    fk_target_field_label,
                    column_label,
                )

        # Empty strings not accepted by Metabase
        column_description = column.description or None
        column_display_name = column.display_name or None
        column_visibility = column.visibility_type or "normal"

        # Preserve this relationship by default
        if api_field["fk_target_field_id"] and not fk_target_field_id:
            fk_target_field_id = api_field["fk_target_field_id"]

        body_field: MutableMapping[str, Optional[Any]] = {}

        # Update if specified, otherwise reset one that had been set
        api_display_name = api_field.get("display_name")
        if api_display_name != column_display_name and (
            column_display_name
            or safe_name(api_display_name) != safe_name(api_field.get("name"))
        ):
            body_field["display_name"] = column_display_name

        if api_field.get("description") != column_description:
            body_field["description"] = column_description
        if api_field.get("visibility_type") != column_visibility:
            body_field["visibility_type"] = column_visibility
        if api_field.get("fk_target_field_id") != fk_target_field_id:
            body_field["fk_target_field_id"] = fk_target_field_id
        if (
            api_field.get("has_field_values") != column.has_field_values
            and column.has_field_values
        ):
            body_field["has_field_values"] = column.has_field_values
        if (
            api_field.get("coercion_strategy") != column.coercion_strategy
            and column.coercion_strategy
        ):
            body_field["coercion_strategy"] = column.coercion_strategy

        settings = api_field.get("settings") or {}
        if settings.get("number_style") != column.number_style and column.number_style:
            settings["number_style"] = column.number_style

        if settings:
            body_field["settings"] = settings

        # Allow explicit null type to override detected one
        api_semantic_type = api_field.get(semantic_type_key)
        if (column.semantic_type and api_semantic_type != column.semantic_type) or (
            column.semantic_type is NullValue and api_semantic_type
        ):
            body_field[semantic_type_key] = column.semantic_type or None

        if body_field:
            ctx.update(entity=api_field, change=body_field, label=column_label)
            _logger.info(
                "Field '%s' will be updated: %s", column_label, ", ".join(body_field)
            )
        else:
            _logger.info("Field '%s' is up to date", column_label)

        return success

    def __get_tables(self, database_id: str) -> Mapping[str, MutableMapping]:
        tables = {}

        metadata = self.metabase.get_database_metadata(database_id)

        bigquery_schema = metadata.get("details", {}).get("dataset-id")

        for table in metadata.get("tables", []):
            # table[schema] is null for bigquery datasets
            table["schema"] = (
                table.get("schema") or bigquery_schema or DEFAULT_SCHEMA
            ).upper()

            fields = {}
            for field in table.get("fields", []):
                new_field = field.copy()
                new_field["kind"] = "field"

                field_name = field["name"].upper()
                fields[field_name] = new_field

            new_table = table.copy()
            new_table["kind"] = "table"
            new_table["fields"] = fields

            schema_name = table["schema"].upper()
            table_name = table["name"].upper()
            tables[f"{schema_name}.{table_name}"] = new_table

        return tables

    @staticmethod
    def __filtered_models(
        models: Iterable[Model],
        database_filter: Optional[Filter],
        schema_filter: Optional[Filter],
        model_filter: Optional[Filter],
        skip_sources: bool,
    ) -> Iterable[Model]:
        def selected(m: Model) -> bool:
            return (
                (not skip_sources or m.group != Group.sources)
                and (not database_filter or database_filter.match(m.database))
                and (not schema_filter or schema_filter.match(m.schema))
                and (not model_filter or model_filter.match(m.name))
            )

        return list(filter(selected, models))

    @dc.dataclass
    class __Context:
        tables: Mapping[str, MutableMapping] = dc.field(default_factory=dict)
        updates: MutableMapping[str, MutableMapping] = dc.field(default_factory=dict)

        def get_field(self, table_key: str, field_key: str) -> MutableMapping:
            return self.tables.get(table_key, {}).get("fields", {}).get(field_key, {})

        def update(self, entity: MutableMapping, change: Mapping, label: str):
            entity.update(change)

            key = f"{entity['kind']}.{entity['id']}"
            update = self.updates.get(key, {})
            update["kind"] = entity["kind"]
            update["id"] = entity["id"]
            update["label"] = label

            body = update.get("body", {})
            body.update(change)
            update["body"] = body

            self.updates[key] = update



import logging

from dbtmetabase.format import setup_logging

from .test_exposures import *
from .test_format import *
from .test_manifest import *
from .test_metabase import *
from .test_models import *

setup_logging(level=logging.DEBUG, path=None)



import functools
import logging
from pathlib import Path
from typing import Any, Callable, List, Mapping, Optional, Sequence, Tuple, Union, cast

import click
import yaml

from .core import DbtMetabase
from .format import Filter, setup_logging


def _click_list_option_kwargs() -> Mapping[str, Any]:
    """Click option that accepts comma-separated values.

    Built-in list only allows repeated flags, which is ugly for larger lists.

    Returns:
        Mapping[str, Any]: Mapping of kwargs (to be unpacked with **).
    """

    def callback(
        ctx: click.Context,
        param: click.Option,
        value: Union[str, List[str]],
    ) -> Optional[List[str]]:
        if value is None:
            return None

        if ctx.get_parameter_source(str(param.name)) in (
            click.core.ParameterSource.DEFAULT,
            click.core.ParameterSource.DEFAULT_MAP,
        ) and isinstance(value, list):
            # Lists in defaults (config or option) should be lists
            return value

        if isinstance(value, str):
            str_value = value
        elif isinstance(value, list):
            # When type=list, string value will be a list of chars
            str_value = "".join(value)
        else:
            raise click.BadParameter("must be comma-separated list")

        return str_value.split(",")

    return {
        "type": click.UNPROCESSED,
        "callback": callback,
    }


@click.group()
@click.version_option(package_name="dbt-metabase")
@click.option(
    "--config-path",
    default="~/.dbt-metabase/config.yml",
    show_default=True,
    type=click.Path(),
    help="Path to config.yml file with default values.",
)
@click.pass_context
def cli(ctx: click.Context, config_path: str):
    group = cast(click.Group, ctx.command)

    config_path_expanded = Path(config_path).expanduser()
    if config_path_expanded.exists():
        with open(config_path_expanded, "r", encoding="utf-8") as f:
            config = yaml.safe_load(f).get("config", {})
            # Propagate common configs to all commands
            common = {k: v for k, v in config.items() if k not in group.commands}
            ctx.default_map = {
                command: {**common, **config.get(command, {})}
                for command in group.commands
            }


def _add_setup(func: Callable) -> Callable:
    """Add common options and initialize core."""

    @click.option(
        "--manifest-path",
        envvar="MANIFEST_PATH",
        show_envvar=True,
        required=True,
        type=click.Path(exists=True, dir_okay=False),
        help="Path to dbt manifest.json, usually in target/ directory after compilation.",
    )
    @click.option(
        "--metabase-url",
        metavar="URL",
        envvar="METABASE_URL",
        show_envvar=True,
        required=True,
        type=click.STRING,
        help="Metabase URL, e.g. 'https://metabase.example.com'.",
    )
    @click.option(
        "--metabase-username",
        metavar="USERNAME",
        envvar="METABASE_USERNAME",
        show_envvar=True,
        type=click.STRING,
        help="Metabase username (required unless providing session ID).",
    )
    @click.option(
        "--metabase-password",
        metavar="PASSWORD",
        envvar="METABASE_PASSWORD",
        show_envvar=True,
        type=click.STRING,
        help="Metabase password (required unless providing session ID).",
    )
    @click.option(
        "--metabase-session-id",
        metavar="TOKEN",
        envvar="METABASE_SESSION_ID",
        show_envvar=True,
        type=click.STRING,
        help="Metabase session ID (alternative to username/password).",
    )
    @click.option(
        "--skip-verify",
        envvar="SKIP_VERIFY",
        show_envvar=True,
        help="Skip TLS certificate verification (not recommended).",
    )
    @click.option(
        "--cert",
        metavar="CERT",
        envvar="CERT",
        show_envvar=True,
        type=click.Path(exists=True, dir_okay=False),
        help="Path to TLS certificate bundle.",
    )
    @click.option(
        "--http-timeout",
        metavar="SECS",
        envvar="HTTP_TIMEOUT",
        show_envvar=True,
        type=click.INT,
        default=DbtMetabase.DEFAULT_HTTP_TIMEOUT,
        show_default=True,
        help="HTTP timeout in seconds.",
    )
    @click.option(
        "--http-header",
        "http_headers",
        type=(str, str),
        multiple=True,
        help="Additional HTTP request headers.",
    )
    @click.option(
        "-v",
        "--verbose",
        is_flag=True,
        help="Enable verbose logging.",
    )
    @functools.wraps(func)
    def wrapper(
        manifest_path: str,
        metabase_url: str,
        metabase_username: str,
        metabase_password: str,
        metabase_session_id: Optional[str],
        skip_verify: bool,
        cert: Optional[str],
        http_timeout: int,
        http_headers: Sequence[Tuple[str, str]],
        verbose: bool,
        **kwargs,
    ):
        setup_logging(
            level=logging.DEBUG if verbose else logging.INFO,
            path=Path.home().absolute() / ".dbt-metabase" / "logs" / "dbtmetabase.log",
        )

        return func(
            core=DbtMetabase(
                manifest_path=manifest_path,
                metabase_url=metabase_url,
                metabase_username=metabase_username,
                metabase_password=metabase_password,
                metabase_session_id=metabase_session_id,
                skip_verify=skip_verify,
                cert=cert,
                http_timeout=http_timeout,
                http_headers={k: v for k, v in http_headers},
            ),
            **kwargs,
        )

    return wrapper


@cli.command(help="Export dbt models to Metabase.")
@_add_setup
@click.option(
    "--metabase-database",
    metavar="METABASE_DATABASE",
    envvar="METABASE_DATABASE",
    show_envvar=True,
    required=True,
    type=click.STRING,
    help="Target database in Metabase.",
)
@click.option(
    "--include-databases",
    metavar="DATABASES",
    envvar="INCLUDE_DATABASES",
    show_envvar=True,
    **_click_list_option_kwargs(),
    help="Only include certain dbt databases.",
)
@click.option(
    "--exclude-databases",
    metavar="DATABASES",
    envvar="EXCLUDE_DATABASES",
    show_envvar=True,
    **_click_list_option_kwargs(),
    help="Exclude certain dbt databases.",
)
@click.option(
    "--include-schemas",
    metavar="SCHEMAS",
    envvar="INCLUDE_SCHEMAS",
    show_envvar=True,
    **_click_list_option_kwargs(),
    help="Only include certain dbt schemas.",
)
@click.option(
    "--exclude-schemas",
    metavar="SCHEMAS",
    envvar="EXCLUDE_SCHEMAS",
    show_envvar=True,
    **_click_list_option_kwargs(),
    help="Exclude certain dbt schemas.",
)
@click.option(
    "--include-models",
    metavar="MODELS",
    envvar="INCLUDE_MODELS",
    show_envvar=True,
    **_click_list_option_kwargs(),
    help="Only include certain dbt models.",
)
@click.option(
    "--exclude-models",
    metavar="MODELS",
    envvar="EXCLUDE_MODELS",
    show_envvar=True,
    **_click_list_option_kwargs(),
    help="Exclude certain dbt models.",
)
@click.option(
    "--sync-timeout",
    metavar="SECS",
    envvar="SYNC_TIMEOUT",
    show_envvar=True,
    default=DbtMetabase.DEFAULT_MODELS_SYNC_TIMEOUT,
    type=click.INT,
    help="Number of seconds to wait until Metabase schema matches the dbt project. To skip synchronization, set timeout to 0.",
)
@click.option(
    "--skip-sources",
    envvar="SKIP_SOURCES",
    show_envvar=True,
    is_flag=True,
    help="Exclude dbt sources from export.",
)
@click.option(
    "--append-tags",
    envvar="APPEND_TAGS",
    show_envvar=True,
    is_flag=True,
    help="Append dbt tags to table descriptions.",
)
@click.option(
    "--docs-url",
    metavar="URL",
    envvar="DOCS_URL",
    show_envvar=True,
    type=click.STRING,
    help="URL for dbt docs hosting, to append model links to table descriptions.",
)
@click.option(
    "--order-fields",
    envvar="ORDER_FIELDS",
    show_envvar=True,
    is_flag=True,
    help="Preserve column order in dbt project.",
)
def models(
    metabase_database: str,
    include_databases: Optional[Sequence[str]],
    exclude_databases: Optional[Sequence[str]],
    include_schemas: Optional[Sequence[str]],
    exclude_schemas: Optional[Sequence[str]],
    include_models: Optional[Sequence[str]],
    exclude_models: Optional[Sequence[str]],
    skip_sources: bool,
    sync_timeout: int,
    append_tags: bool,
    docs_url: Optional[str],
    order_fields: bool,
    core: DbtMetabase,
):
    core.export_models(
        metabase_database=metabase_database,
        database_filter=Filter(include=include_databases, exclude=exclude_databases),
        schema_filter=Filter(include=include_schemas, exclude=exclude_schemas),
        model_filter=Filter(include=include_models, exclude=exclude_models),
        skip_sources=skip_sources,
        sync_timeout=sync_timeout,
        append_tags=append_tags,
        docs_url=docs_url,
        order_fields=order_fields,
    )


@cli.command(help="Extract dbt exposures from Metabase.")
@_add_setup
@click.option(
    "--output-path",
    envvar="OUTPUT_PATH",
    show_envvar=True,
    type=click.Path(exists=True, file_okay=False),
    default=DbtMetabase.DEFAULT_EXPOSURES_OUTPUT_PATH,
    show_default=True,
    help="Output path for exposure YAML files.",
)
@click.option(
    "--output-grouping",
    envvar="OUTPUT_GROUPING",
    show_envvar=True,
    type=click.Choice(["collection", "type"]),
    help="Grouping key for exposure YAML files.",
)
@click.option(
    "--include-collections",
    metavar="COLLECTIONS",
    envvar="INCLUDE_COLLECTIONS",
    show_envvar=True,
    **_click_list_option_kwargs(),
    help="Only include certain Metabase collections.",
)
@click.option(
    "--exclude-collections",
    metavar="COLLECTIONS",
    envvar="EXCLUDE_COLLECTIONS",
    show_envvar=True,
    **_click_list_option_kwargs(),
    help="Exclude certain Metabase collections.",
)
@click.option(
    "--allow-personal-collections",
    envvar="ALLOW_PERSONAL_COLLECTIONS",
    show_envvar=True,
    is_flag=True,
    help="Include personal Metabase collections.",
)
def exposures(
    output_path: str,
    output_grouping: Optional[str],
    include_collections: Optional[Sequence[str]],
    exclude_collections: Optional[Sequence[str]],
    allow_personal_collections: bool,
    core: DbtMetabase,
):
    core.extract_exposures(
        output_path=output_path,
        output_grouping=output_grouping,
        collection_filter=Filter(
            include=include_collections,
            exclude=exclude_collections,
        ),
        allow_personal_collections=allow_personal_collections,
    )


if __name__ == "__main__":
    # Executed when running locally via python3 -m dbtmetabase
    cli()  # pylint: disable=no-value-for-parameter



from __future__ import annotations

import dataclasses as dc
import logging
import re
from abc import ABCMeta, abstractmethod
from operator import itemgetter
from pathlib import Path
from typing import Iterable, Mapping, MutableMapping, MutableSequence, Optional, Tuple

from dbtmetabase.metabase import Metabase

from .errors import ArgumentError
from .format import Filter, dump_yaml, safe_description, safe_name
from .manifest import Manifest

_RESOURCE_VERSION = 2

# Extracting table in `from` and `join` clauses (won't recognize some valid SQL, e.g. `from "table with spaces"`)
_EXPOSURE_PARSER = re.compile(r"[FfJj][RrOo][OoIi][MmNn]\s+([\w.\"]+)")
_CTE_PARSER = re.compile(
    r"[Ww][Ii][Tt][Hh]\s+\b(\w+)\b\s+as|[)]\s*[,]\s*\b(\w+)\b\s+as"
)

_logger = logging.getLogger(__name__)


class ExposuresMixin(metaclass=ABCMeta):
    """Abstraction for extracting exposures."""

    DEFAULT_EXPOSURES_OUTPUT_PATH = "."

    @property
    @abstractmethod
    def manifest(self) -> Manifest:
        pass

    @property
    @abstractmethod
    def metabase(self) -> Metabase:
        pass

    def extract_exposures(
        self,
        output_path: str = DEFAULT_EXPOSURES_OUTPUT_PATH,
        output_grouping: Optional[str] = None,
        collection_filter: Optional[Filter] = None,
        allow_personal_collections: bool = False,
    ) -> Iterable[Mapping]:
        """Extract dbt exposures from Metabase.

        Args:
            output_path (str, optional): Path for output files. Defaults to ".".
            output_grouping (Optional[str], optional): Grouping for output YAML files, supported values: "collection" (by collection slug) or "type" (by entity type). Defaults to None.
            collection_filter (Optional[Filter], optional): Filter Metabase collections. Defaults to None.
            allow_personal_collections (bool, optional): Allow personal Metabase collections. Defaults to False.

        Returns:
            Iterable[Mapping]: List of parsed exposures.
        """

        if output_grouping not in (None, "collection", "type"):
            raise ArgumentError(f"Unsupported grouping: {output_grouping}")

        collection_filter = collection_filter or Filter()

        models = self.manifest.read_models()

        ctx = self.__Context(
            model_refs={m.name: m.ref for m in models if m.ref},
            table_names={t["id"]: t["name"] for t in self.metabase.get_tables()},
        )

        exposures = []
        counts: MutableMapping[str, int] = {}

        for collection in self.metabase.get_collections(
            exclude_personal=not allow_personal_collections
        ):
            collection_name = collection["name"]
            collection_slug = collection.get("slug", safe_name(collection["name"]))

            if not collection_filter.match(collection_name):
                _logger.debug("Skipping collection '%s'", collection["name"])
                continue

            _logger.info("Exploring collection '%s'", collection["name"])
            for item in self.metabase.get_collection_items(
                uid=collection["id"],
                models=("card", "dashboard"),
            ):
                depends = set()
                native_query = ""
                header = ""

                entity: Mapping
                if item["model"] == "card":
                    card_entity = self.metabase.find_card(uid=item["id"])
                    if card_entity is None:
                        _logger.info("Card '%s' not found, skipping", item["id"])
                        continue

                    entity = card_entity
                    header = (
                        f"Visualization: {entity.get('display', 'Unknown').title()}"
                    )

                    result = self.__extract_card_exposures(ctx, card=entity)
                    depends.update(result["depends"])
                    native_query = result["native_query"]

                elif item["model"] == "dashboard":
                    dashboard_entity = self.metabase.find_dashboard(uid=item["id"])
                    if dashboard_entity is None:
                        _logger.info("Dashboard '%s' not found, skipping", item["id"])
                        continue

                    entity = dashboard_entity
                    cards = entity.get("ordered_cards", [])
                    if not cards:
                        continue

                    header = f"Dashboard Cards: {len(cards)}"
                    for card_ref in cards:
                        card = card_ref.get("card", {})
                        if "id" not in card:
                            continue

                        depends.update(
                            self.__extract_card_exposures(
                                ctx,
                                card=self.metabase.find_card(uid=card["id"]),
                            )["depends"]
                        )
                else:
                    _logger.warning("Unexpected collection item '%s'", item["model"])
                    continue

                name = entity.get("name", "Exposure [Unresolved Name]")
                _logger.info("Processing %s '%s'", item["model"], name)

                creator_name = None
                creator_email = None
                if "creator" in entity:
                    creator_email = entity["creator"]["email"]
                    creator_name = entity["creator"]["common_name"]
                elif "creator_id" in entity:
                    creator = self.metabase.find_user(uid=entity["creator_id"])
                    if creator:
                        creator_name = creator.get("common_name")
                        creator_email = creator.get("email")

                label = name
                name = safe_name(name)
                count = counts.get(name, 0)
                counts[name] = count + 1

                exposures.append(
                    {
                        "id": item["id"],
                        "type": item["model"],
                        "collection": collection_slug,
                        "body": self.__format_exposure(
                            model=item["model"],
                            uid=item["id"],
                            name=name + (f"_{count}" if count > 0 else ""),
                            label=label,
                            header=header,
                            description=entity.get("description", ""),
                            created_at=entity["created_at"],
                            creator_name=creator_name or "",
                            creator_email=creator_email or "",
                            native_query=native_query,
                            depends_on={
                                ctx.model_refs[depend]
                                for depend in depends
                                if depend in ctx.model_refs
                            },
                        ),
                    }
                )

        self.__write_exposures(exposures, output_path, output_grouping)

        return exposures

    def __extract_card_exposures(
        self,
        ctx: __Context,
        card: Optional[Mapping],
    ) -> Mapping:
        """Extracts exposures from Metabase questions."""

        depends = set()
        native_query = ""

        if card:
            query = card.get("dataset_query", {})
            if query.get("type") == "query":
                # Metabase GUI derived query
                query_source = query.get("query", {}).get(
                    "source-table", card.get("table_id")
                )

                if str(query_source).startswith("card__"):
                    # Handle questions based on other questions
                    depends.update(
                        self.__extract_card_exposures(
                            ctx,
                            card=self.metabase.find_card(
                                uid=query_source.split("__")[-1]
                            ),
                        )["depends"]
                    )
                elif query_source in ctx.table_names:
                    # Normal question
                    source_table = ctx.table_names.get(query_source)
                    if source_table:
                        source_table = source_table.lower()
                        _logger.info("Extracted model '%s' from card", source_table)
                        depends.add(source_table)

                # Find models exposed through joins
                for join in query.get("query", {}).get("joins", []):
                    join_source = join.get("source-table")

                    if str(join_source).startswith("card__"):
                        # Handle questions based on other questions
                        depends.update(
                            self.__extract_card_exposures(
                                ctx,
                                card=self.metabase.find_card(
                                    uid=join_source.split("__")[-1]
                                ),
                            )["depends"]
                        )
                        continue

                    # Joined model parsed
                    joined_table = ctx.table_names.get(join_source)
                    if joined_table:
                        joined_table = joined_table.lower()
                        _logger.info("Extracted model '%s' from join", joined_table)
                        depends.add(joined_table)

            elif query.get("type") == "native":
                # Metabase native query
                native_query = query["native"].get("query")
                ctes: MutableSequence[str] = []

                # Parse common table expressions for exclusion
                for matched_cte in re.findall(_CTE_PARSER, native_query):
                    ctes.extend(group.lower() for group in matched_cte if group)

                # Parse SQL for exposures through FROM or JOIN clauses
                for sql_ref in re.findall(_EXPOSURE_PARSER, native_query):
                    # Grab just the table / model name
                    parsed_model = sql_ref.split(".")[-1].strip('"').lower()

                    # Scrub CTEs (qualified sql_refs can not reference CTEs)
                    if parsed_model in ctes and "." not in sql_ref:
                        continue

                    # Verify this is one of our parsed refable models so exposures dont break the DAG
                    if not ctx.model_refs.get(parsed_model):
                        continue

                    if parsed_model:
                        _logger.info(
                            "Extracted model '%s' from native query", parsed_model
                        )
                        depends.add(parsed_model)

        return {
            "depends": depends,
            "native_query": native_query,
        }

    def __format_exposure(
        self,
        model: str,
        uid: str,
        name: str,
        label: str,
        header: str,
        description: str,
        created_at: str,
        creator_name: str,
        creator_email: str,
        native_query: Optional[str],
        depends_on: Iterable[str],
    ) -> Mapping:
        """Builds dbt exposure representation (see https://docs.getdbt.com/reference/exposure-properties)."""

        dbt_type: str
        url: str
        if model == "card":
            dbt_type = "analysis"
            url = self.metabase.format_card_url(uid=uid)
        elif model == "dashboard":
            dbt_type = "dashboard"
            url = self.metabase.format_dashboard_url(uid=uid)
        else:
            raise ValueError("Unexpected exposure type")

        if header:
            header = f"### {header}\n\n"

        if description:
            description = description.strip()
        else:
            description = "No description provided in Metabase"

        if native_query:
            # Format query into markdown code block
            native_query = "\n".join(l for l in native_query.split("\n") if l.strip())
            native_query = f"#### Query\n\n```\n{native_query}\n```\n\n"

        metadata = (
            "#### Metadata\n\n"
            + f"Metabase ID: __{uid}__\n\n"
            + f"Created On: __{created_at}__"
        )

        return {
            "name": name,
            "label": label,
            "description": safe_description(
                f"{header}{description}\n\n{native_query}{metadata}"
            ),
            "type": dbt_type,
            "url": url,
            "maturity": "medium",
            "owner": {
                "name": creator_name,
                "email": creator_email,
            },
            "depends_on": list(depends_on),
        }

    @staticmethod
    def __write_exposures(
        exposures: Iterable[Mapping],
        output_path: str,
        output_grouping: Optional[str],
    ):
        """Write exposures to output files."""

        grouped: MutableMapping[Tuple[str, ...], MutableSequence[Mapping]] = {}
        for exposure in exposures:
            group: Tuple[str, ...] = ("exposures",)
            if output_grouping == "collection":
                group = (exposure["collection"],)
            elif output_grouping == "type":
                group = (exposure["type"], exposure["id"])

            exps = grouped.get(group, [])
            exps.append(exposure)
            if group not in grouped:
                grouped[group] = exps

        for group, exps in grouped.items():
            path = Path(output_path).expanduser()
            path = path.joinpath(*group[:-1]) / f"{group[-1]}.yml"
            path.parent.mkdir(parents=True, exist_ok=True)

            exps_unwrapped = map(lambda x: x["body"], exps)
            exps_sorted = sorted(exps_unwrapped, key=itemgetter("name"))

            with open(path, "w", encoding="utf-8") as f:
                dump_yaml(
                    data={
                        "version": _RESOURCE_VERSION,
                        "exposures": exps_sorted,
                    },
                    stream=f,
                )

    @dc.dataclass
    class __Context:
        model_refs: Mapping[str, str]
        table_names: Mapping[str, str]



#!/usr/bin/env python
import time

import requests
from molot import envarg, envarg_int, evaluate, shell, target

POSTGRES_HOST = envarg("POSTGRES_HOST")
POSTGRES_PORT = envarg_int("POSTGRES_PORT")
POSTGRES_DB = envarg("POSTGRES_DB")
POSTGRES_USER = envarg("POSTGRES_USER")
POSTGRES_PASSWORD = envarg("POSTGRES_PASSWORD")
MB_HOST = envarg("MB_HOST")
MB_PORT = envarg_int("MB_PORT")
MB_SETUP_TOKEN = envarg("MB_SETUP_TOKEN")
MB_USER = envarg("MB_USER")
MB_PASSWORD = envarg("MB_PASSWORD")
MB_NAME = envarg("MB_NAME", "dbtmetabase")


@target(
    description="initial setup",
    depends=["metabase_setup", "dbt_run", "metabase_content"],
)
def init():
    pass


@target(description="run dbt project")
def dbt_run():
    shell("dbt seed --profiles-dir .")
    shell("dbt run --profiles-dir .")


@target(description="set up Metabase user and database")
def metabase_setup():
    requests.post(
        url=f"http://{MB_HOST}:{MB_PORT}/api/setup",
        json={
            "token": MB_SETUP_TOKEN,
            "user": {
                "site_name": MB_NAME,
                "first_name": MB_NAME,
                "last_name": None,
                "email": MB_USER,
                "password_confirm": MB_PASSWORD,
                "password": MB_PASSWORD,
            },
            "database": {
                "engine": "postgres",
                "name": POSTGRES_DB,
                "details": {
                    "host": POSTGRES_HOST,
                    "port": POSTGRES_PORT,
                    "dbname": POSTGRES_DB,
                    "user": POSTGRES_USER,
                    "password": POSTGRES_PASSWORD,
                    "schema-filters-type": "all",
                    "ssl": False,
                    "tunnel-enabled": False,
                    "advanced-options": False,
                },
                "is_on_demand": False,
                "is_full_sync": True,
                "is_sample": False,
                "cache_ttl": None,
                "refingerprint": False,
                "auto_run_queries": True,
                "schedules": {},
            },
            "prefs": {
                "site_name": MB_NAME,
                "site_locale": "en",
                "allow_tracking": "false",
            },
        },
        timeout=10,
    ).raise_for_status()


@target(description="add mock content to Metabase")
def metabase_content():
    session_id = requests.post(
        url=f"http://{MB_HOST}:{MB_PORT}/api/session",
        json={"username": MB_USER, "password": MB_PASSWORD},
        timeout=10,
    ).json()["id"]

    headers = {"X-Metabase-Session": session_id}

    database_id = ""
    databases = requests.get(
        url=f"http://{MB_HOST}:{MB_PORT}/api/database",
        headers=headers,
        json={},
        timeout=10,
    ).json()["data"]
    for db in databases:
        if db["name"] == POSTGRES_DB:
            database_id = db["id"]
            break

    requests.post(
        url=f"http://{MB_HOST}:{MB_PORT}/api/database/{database_id}/sync_schema",
        headers=headers,
        json={},
        timeout=10,
    )

    time.sleep(5)

    tables_fields = requests.get(
        url=f"http://{MB_HOST}:{MB_PORT}/api/database/{database_id}?include=tables.fields",
        headers=headers,
        timeout=10,
    ).json()

    customers_table_id = ""
    first_order_field_id = ""
    for table in tables_fields["tables"]:
        if table["name"] == "customers":
            customers_table_id = table["id"]
            for field in table["fields"]:
                if field["name"] == "first_order":
                    first_order_field_id = field["id"]
                    break

    requests.post(
        url=f"http://{MB_HOST}:{MB_PORT}/api/card",
        headers=headers,
        json={
            "name": "Customers",
            "dataset_query": {
                "database": database_id,
                "type": "query",
                "query": {
                    "source-table": customers_table_id,
                    "aggregation": [["count"]],
                    "breakout": [
                        [
                            "field",
                            first_order_field_id,
                            {"base-type": "type/Date", "temporal-unit": "month"},
                        ]
                    ],
                },
            },
            "display": "line",
            "description": "Customers test",
            "visualization_settings": {
                "graph.dimensions": ["first_order"],
                "graph.metrics": ["count"],
            },
            "collection_id": None,
            "collection_position": None,
            "result_metadata": [
                {
                    "description": None,
                    "semantic_type": None,
                    "coercion_strategy": None,
                    "unit": "month",
                    "name": "first_order",
                    "settings": None,
                    "fk_target_field_id": None,
                    "field_ref": [
                        "field",
                        first_order_field_id,
                        {"base-type": "type/Date", "temporal-unit": "month"},
                    ],
                    "effective_type": "type/DateTimeWithLocalTZ",
                    "id": first_order_field_id,
                    "visibility_type": "normal",
                    "display_name": "First Order",
                    "fingerprint": {
                        "global": {"distinct-count": 47, "nil%": 0.38},
                        "type": {
                            "type/DateTime": {
                                "earliest": "2018-01-01",
                                "latest": "2018-04-07",
                            }
                        },
                    },
                    "base_type": "type/DateTimeWithLocalTZ",
                },
                {
                    "display_name": "Count",
                    "semantic_type": "type/Quantity",
                    "field_ref": ["aggregation", 0],
                    "name": "count",
                    "base_type": "type/BigInteger",
                    "effective_type": "type/BigInteger",
                    "fingerprint": {
                        "global": {"distinct-count": 4, "nil%": 0},
                        "type": {
                            "type/Number": {
                                "min": 2,
                                "q1": 11.298221281347036,
                                "q3": 27.5,
                                "max": 38,
                                "sd": 12.96148139681572,
                                "avg": 20,
                            }
                        },
                    },
                },
            ],
        },
        timeout=10,
    )


evaluate()



import json
from pathlib import Path
from typing import Any, Dict, Mapping, Optional, Sequence

import requests

from dbtmetabase.core import DbtMetabase
from dbtmetabase.manifest import Manifest, Model
from dbtmetabase.metabase import Metabase

FIXTURES_PATH = Path("tests") / "fixtures"
TMP_PATH = Path("tests") / "tmp"


class MockMetabase(Metabase):
    def __init__(self, url: str):
        super().__init__(
            url=url,
            username=None,
            password=None,
            session_id="dummy",
            skip_verify=False,
            cert=None,
            http_timeout=1,
            http_headers=None,
            http_adapter=None,
        )

    def _api(
        self,
        method: str,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Mapping:
        path_toks = f"{path.lstrip('/')}.json".split("/")
        if path_toks[0] == "api" and method == "get":
            json_path = Path.joinpath(FIXTURES_PATH, *path_toks)
            if json_path.exists():
                with open(json_path, encoding="utf-8") as f:
                    return json.load(f)
            else:
                response = requests.Response()
                response.status_code = 404
                raise requests.exceptions.HTTPError(response=response)
        return {}


class MockManifest(Manifest):
    _models: Sequence[Model] = []

    def read_models(self) -> Sequence[Model]:
        if not self._models:
            self._models = super().read_models()
        return self._models


class MockDbtMetabase(DbtMetabase):
    def __init__(
        self,
        manifest_path: Path = FIXTURES_PATH / "manifest-v2.json",
        metabase_url: str = "http://localhost:3000",
    ):  # pylint: disable=super-init-not-called
        self._manifest = MockManifest(path=manifest_path)
        self._metabase = MockMetabase(url=metabase_url)



import unittest

from ._mocks import MockMetabase


class TestMetabase(unittest.TestCase):
    def setUp(self):
        self.metabase = MockMetabase(url="http://localhost")

    def test_metabase_find_database(self):
        db = self.metabase.find_database(name="unit_testing")
        assert db
        self.assertEqual(2, db["id"])
        self.assertIsNone(self.metabase.find_database(name="foo"))

    def test_metabase_get_collections(self):
        excluded = self.metabase.get_collections(exclude_personal=True)
        self.assertEqual(3, len(excluded))

        included = self.metabase.get_collections(exclude_personal=False)
        self.assertEqual(4, len(included))

    def test_metabase_get_collection_items(self):
        cards = self.metabase.get_collection_items(
            uid="3",
            models=("card",),
        )
        self.assertEqual({"card"}, {item["model"] for item in cards})

        dashboards = self.metabase.get_collection_items(
            uid="3",
            models=("dashboard",),
        )
        self.assertEqual({"dashboard"}, {item["model"] for item in dashboards})

        both = self.metabase.get_collection_items(
            uid="3",
            models=("card", "dashboard"),
        )
        self.assertEqual({"card", "dashboard"}, {item["model"] for item in both})



from __future__ import annotations

import logging
from pathlib import Path
from typing import Optional, Tuple, Union

from requests.adapters import HTTPAdapter

from ._exposures import ExposuresMixin
from ._models import ModelsMixin
from .manifest import Manifest
from .metabase import Metabase

_logger = logging.getLogger(__name__)


class DbtMetabase(ModelsMixin, ExposuresMixin):
    """dbt + Metabase integration."""

    DEFAULT_HTTP_TIMEOUT = 15

    def __init__(
        self,
        manifest_path: Union[str, Path],
        metabase_url: str,
        metabase_username: Optional[str] = None,
        metabase_password: Optional[str] = None,
        metabase_session_id: Optional[str] = None,
        skip_verify: bool = False,
        cert: Optional[Union[str, Tuple[str, str]]] = None,
        http_timeout: int = DEFAULT_HTTP_TIMEOUT,
        http_headers: Optional[dict] = None,
        http_adapter: Optional[HTTPAdapter] = None,
    ):
        """dbt + Metabase integration.

        Args:
            manifest_path (Union[str,Path]): Path to dbt manifest.json, usually in target/ directory after compilation.
            metabase_url (str): Metabase URL, e.g. "https://metabase.example.com".
            metabase_username (Optional[str], optional): Metabase username (required unless providing session ID). Defaults to None.
            metabase_password (Optional[str], optional): Metabase password (required unless providing session ID). Defaults to None.
            metabase_session_id (Optional[str], optional): Metabase session ID. Defaults to None.
            skip_verify (bool, optional): Skip TLS certificate verification (not recommended). Defaults to False.
            cert (Optional[Union[str, Tuple[str, str]]], optional): Path to a custom certificate. Defaults to None.
            http_timeout (int, optional): HTTP request timeout in secs. Defaults to 15.
            http_headers (Optional[dict], optional): Additional HTTP headers. Defaults to None.
            http_adapter (Optional[HTTPAdapter], optional): Custom requests HTTP adapter. Defaults to None.
        """

        self._manifest = Manifest(
            path=manifest_path,
        )
        self._metabase = Metabase(
            url=metabase_url,
            username=metabase_username,
            password=metabase_password,
            session_id=metabase_session_id,
            skip_verify=skip_verify,
            cert=cert,
            http_timeout=http_timeout,
            http_headers=http_headers,
            http_adapter=http_adapter,
        )

    @property
    def manifest(self) -> Manifest:
        return self._manifest

    @property
    def metabase(self) -> Metabase:
        return self._metabase



from __future__ import annotations

import dataclasses as dc
import json
import logging
import re
from enum import Enum
from pathlib import Path
from typing import (
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Union,
)

from .format import NullValue

_logger = logging.getLogger(__name__)

# Namespace for meta fields, e.g. metabase.field
_META_NS = "metabase"
# Allowed namespace fields
_COMMON_META_FIELDS = [
    "display_name",
    "visibility_type",
]
# Must be covered by Column attributes
_COLUMN_META_FIELDS = _COMMON_META_FIELDS + [
    "semantic_type",
    "has_field_values",
    "coercion_strate`gy",
    "number_style",
]
# Must be covered by Model attributes
_MODEL_META_FIELDS = _COMMON_META_FIELDS + [
    "points_of_interest",
    "caveats",
]

# Default model schema (only schema in BigQuery)
DEFAULT_SCHEMA = "PUBLIC"

# Foreign key constraint: "schema.model (column)" / "model (column)"
_CONSTRAINT_FK_PARSER = re.compile(r"(?P<model>.+)\s+\((?P<column>.+)\)")


class Manifest:
    """dbt manifest reader."""

    def __init__(self, path: Union[str, Path]):
        """Reader for compiled dbt manifest.json file.

        Args:
            path (Union[str, Path]): Path to dbt manifest.json (usually under target/).
        """

        self.path = Path(path).expanduser()

    def read_models(self) -> Sequence[Model]:
        """Reads dbt models in Metabase-friendly format.

        Returns:
            Sequence[Model]: List of dbt models in Metabase-friendly format.
        """

        with open(self.path, "r", encoding="utf-8") as f:
            manifest = json.load(f)

        models: MutableSequence[Model] = []

        for node in manifest["nodes"].values():
            if node["resource_type"] != "model":
                continue

            name = node["name"]
            if node["config"]["materialized"] == "ephemeral":
                _logger.debug("Skipping ephemeral model '%s'", name)
                continue

            models.append(self._read_model(manifest, node, Group.nodes))

        for node in manifest["sources"].values():
            if node["resource_type"] != "source":
                continue

            models.append(
                self._read_model(manifest, node, Group.sources, node["source_name"])
            )

        return models

    def _read_model(
        self,
        manifest: Mapping,
        manifest_model: Mapping,
        group: Group,
        source: Optional[str] = None,
    ) -> Model:
        database = manifest_model["database"]
        schema = manifest_model["schema"]
        unique_id = manifest_model["unique_id"]

        relationships = self._read_relationships(manifest, group, unique_id)

        columns = [
            self._read_column(column, schema, relationships.get(column["name"]))
            for column in manifest_model.get("columns", {}).values()
        ]

        return Model(
            database=database,
            schema=schema,
            group=group,
            name=manifest_model["name"],
            alias=manifest_model.get(
                "alias", manifest_model.get("identifier", manifest_model["name"])
            ),
            description=manifest_model.get("description"),
            columns=columns,
            unique_id=unique_id,
            source=source,
            tags=manifest_model.get("tags", []),
            **self._scan_fields(
                manifest_model.get("meta", {}),
                fields=_MODEL_META_FIELDS,
                ns=_META_NS,
            ),
        )

    def _read_column(
        self,
        manifest_column: Mapping,
        schema: str,
        relationship: Optional[Mapping],
    ) -> Column:
        column = Column(
            name=manifest_column.get("name", ""),
            description=manifest_column.get("description"),
            **self._scan_fields(
                manifest_column.get("meta", {}),
                fields=_COLUMN_META_FIELDS,
                ns=_META_NS,
            ),
        )

        self._set_column_relationship(
            manifest_column=manifest_column,
            column=column,
            schema=schema,
            relationship=relationship,
        )

        return column

    def _read_relationships(
        self,
        manifest: Mapping,
        group: Group,
        unique_id: str,
    ) -> Mapping[str, Mapping[str, str]]:
        relationships = {}

        for child_id in manifest["child_map"][unique_id]:
            child = manifest.get(group, {}).get(child_id, {})
            child_name = child.get("alias", child.get("name"))

            if (
                child.get("resource_type") == "test"
                and child.get("test_metadata", {}).get("name") == "relationships"
            ):
                # To get the name of the foreign table, we could use child[test_metadata][kwargs][to], which
                # would return the ref() written in the test, but if the model has an alias, that's not enough.
                # Using child[depends_on][nodes] and excluding the current model is better.

                # Nodes contain at most two tables: referenced model and current model (optional).
                depends_on_nodes = list(child["depends_on"][group])

                # Relationships on disabled models mention them in refs but not depends_on,
                # which confuses the filtering logic that follows.
                depends_on_names = {n.split(".")[-1] for n in depends_on_nodes}
                mismatched_refs = []
                for ref in child["refs"]:
                    ref_name = ""
                    if isinstance(ref, dict):  # current manifest
                        ref_name = ref["name"]
                    elif isinstance(ref, list):  # old manifest
                        ref_name = ref[0]
                    if ref_name not in depends_on_names:
                        mismatched_refs.append(ref_name)

                if mismatched_refs:
                    _logger.debug(
                        "Mismatched refs %s with depends_on for relationship '%s', skipping",
                        mismatched_refs,
                        child_name,
                    )
                    continue

                if len(depends_on_nodes) > 2:
                    _logger.warning(
                        "Unexpected %d depends_on for relationship '%s' instead of <=2, skipping",
                        len(depends_on_nodes),
                        child_name,
                    )
                    continue

                # Skip the incoming relationship tests, in which the fk_target_table is the model currently being read.
                # Otherwise, the primary key of the current model would be (incorrectly) determined to be FK.
                if len(depends_on_nodes) == 2 and depends_on_nodes[1] != unique_id:
                    _logger.debug(
                        "Circular dependency '%s' for relationship '%s', skipping",
                        depends_on_nodes[1],
                        child_name,
                    )
                    continue

                # Remove the current model from the list, ensuring it works for self-referencing models.
                if len(depends_on_nodes) == 2 and unique_id in depends_on_nodes:
                    depends_on_nodes.remove(unique_id)

                if len(depends_on_nodes) != 1:
                    _logger.warning(
                        "Got %d dependencies for '%s' instead of 1, skipping",
                        len(depends_on_nodes),
                        unique_id,
                    )
                    continue

                depends_on_id = depends_on_nodes[0]

                fk_target_model = manifest[group].get(depends_on_id, {})
                fk_target_table = fk_target_model.get(
                    "alias",
                    fk_target_model.get("identifier", fk_target_model.get("name")),
                )
                if not fk_target_table:
                    _logger.debug("Cannot resolve dependency for '%s'", depends_on_id)
                    continue

                fk_target_schema = manifest[group][depends_on_id].get(
                    "schema", DEFAULT_SCHEMA
                )
                fk_target_table = f"{fk_target_schema}.{fk_target_table}"
                fk_target_field = child["test_metadata"]["kwargs"]["field"].strip('"')

                relationships[child["column_name"]] = {
                    "fk_target_table": fk_target_table,
                    "fk_target_field": fk_target_field,
                }

        return relationships

    def _set_column_relationship(
        self,
        manifest_column: Mapping,
        column: Column,
        schema: str,
        relationship: Optional[Mapping],
    ):
        """Sets primary key and foreign key target on a column from constraints, meta fields or provided test relationship."""

        fk_target_table = ""
        fk_target_field = ""

        # Precedence 1: Relationship test
        if relationship:
            fk_target_table = relationship["fk_target_table"]
            fk_target_field = relationship["fk_target_field"]

        # Precedence 2: Constraints
        for constraint in manifest_column.get("constraints", []):
            if constraint["type"] == "primary_key":
                if not column.semantic_type:
                    column.semantic_type = "type/PK"

            elif constraint["type"] == "foreign_key":
                constraint_expr = constraint.get("expression", "")
                constraint_fk = _CONSTRAINT_FK_PARSER.search(constraint_expr)
                if constraint_fk:
                    fk_target_table = constraint_fk.group("model")
                    fk_target_field = constraint_fk.group("column")
                else:
                    _logger.warning(
                        "Unparsable '%s' foreign key constraint: %s",
                        column.name,
                        constraint_expr,
                    )

        # Precedence 3: Meta fields
        meta = manifest_column.get("meta", {})
        fk_target_table = meta.get(f"{_META_NS}.fk_target_table", fk_target_table)
        fk_target_field = meta.get(f"{_META_NS}.fk_target_field", fk_target_field)

        if not fk_target_table or not fk_target_field:
            if fk_target_table or fk_target_table:
                _logger.warning(
                    "Foreign key requires table and field for column '%s'",
                    column.name,
                )
            return

        fk_target_table_path = fk_target_table.split(".")
        if len(fk_target_table_path) == 1 and schema:
            fk_target_table_path.insert(0, schema)

        column.semantic_type = "type/FK"
        column.fk_target_table = ".".join([x.strip('"') for x in fk_target_table_path])
        column.fk_target_field = fk_target_field.strip('"')
        _logger.debug(
            "Relation from '%s' to '%s.%s'",
            column.name,
            column.fk_target_table,
            column.fk_target_field,
        )

    @staticmethod
    def _scan_fields(t: Mapping, fields: Iterable[str], ns: str) -> Mapping:
        """Reads meta fields from a schem object.

        Args:
            t (Mapping): Target to scan for fields.
            fields (List): List of fields to accept.
            ns (str): Field namespace (separated by .).

        Returns:
            Mapping: Field values.
        """

        vals = {}
        for field in fields:
            if f"{ns}.{field}" in t:
                value = t[f"{ns}.{field}"]
                vals[field] = value if value is not None else NullValue
        return vals


class Group(str, Enum):
    nodes = "nodes"
    sources = "sources"


@dc.dataclass
class Column:
    name: str
    description: Optional[str] = None
    display_name: Optional[str] = None
    visibility_type: Optional[str] = None
    semantic_type: Optional[str] = None
    has_field_values: Optional[str] = None
    coercion_strategy: Optional[str] = None
    number_style: Optional[str] = None

    fk_target_table: Optional[str] = None
    fk_target_field: Optional[str] = None

    meta_fields: MutableMapping = dc.field(default_factory=dict)


@dc.dataclass
class Model:
    database: str
    schema: str
    group: Group

    name: str
    alias: str
    description: Optional[str] = None
    display_name: Optional[str] = None
    visibility_type: Optional[str] = None
    points_of_interest: Optional[str] = None
    caveats: Optional[str] = None

    unique_id: Optional[str] = None
    source: Optional[str] = None
    tags: Optional[Sequence[str]] = dc.field(default_factory=list)

    columns: Sequence[Column] = dc.field(default_factory=list)

    @property
    def ref(self) -> Optional[str]:
        if self.group == Group.nodes:
            return f"ref('{self.name}')"
        elif self.group == Group.sources:
            return f"source('{self.source}', '{self.name}')"
        return None

    def format_description(
        self,
        append_tags: bool = False,
        docs_url: Optional[str] = None,
    ) -> str:
        """Formats description from available information.

        Args:
            append_tags (bool, optional): True to include dbt model tags. Defaults to False.
            docs_url (Optional[str], optional): Provide docs base URL to include links. Defaults to None.

        Returns:
            str: Formatted description.
        """

        sections = []

        if self.description:
            sections.append(self.description)

        if append_tags and self.tags:
            sections.append(f"Tags: {', '.join(self.tags)}")

        if docs_url:
            sections.append(
                f"dbt docs: {docs_url.rstrip('/')}/#!/model/{self.unique_id}"
            )

        return "\n\n".join(sections)



import logging

from .core import DbtMetabase
from .format import Filter

__all__ = [
    "DbtMetabase",
    "Filter",
]

try:
    from ._version import __version__ as version  # type: ignore

    __version__ = version
except ModuleNotFoundError:
    logging.warning("No _version.py file")
    __version__ = "0.0.0"



import logging
from typing import Any, Dict, Mapping, Optional, Sequence, Tuple, Union

import requests
from requests.adapters import HTTPAdapter, Retry

from .errors import ArgumentError

_logger = logging.getLogger(__name__)


class Metabase:
    def __init__(
        self,
        url: str,
        username: Optional[str],
        password: Optional[str],
        session_id: Optional[str],
        skip_verify: bool,
        cert: Optional[Union[str, Tuple[str, str]]],
        http_timeout: int,
        http_headers: Optional[dict],
        http_adapter: Optional[HTTPAdapter],
    ):
        self.url = url.rstrip("/")

        self.http_timeout = http_timeout

        self.session = requests.Session()
        self.session.verify = not skip_verify
        self.session.cert = cert

        if http_headers:
            self.session.headers.update(http_headers)

        self.session.mount(
            self.url,
            http_adapter or HTTPAdapter(max_retries=Retry(total=3, backoff_factor=1)),
        )

        if not session_id:
            if username and password:
                session = dict(
                    self._api(
                        method="post",
                        path="/api/session",
                        json={"username": username, "password": password},
                    )
                )
                session_id = str(session["id"])
            else:
                raise ArgumentError("Metabase credentials or session ID required")
        self.session.headers["X-Metabase-Session"] = session_id

        _logger.info("Metabase session established")

    def _api(
        self,
        method: str,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Union[Mapping, Sequence]:
        """Raw API call."""

        if params:
            for key, value in params.items():
                if isinstance(value, bool):
                    params[key] = str(value).lower()

        response = self.session.request(
            method=method,
            url=f"{self.url}{path}",
            params=params,
            timeout=self.http_timeout,
            **kwargs,
        )

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            _logger.error("HTTP request failed: %s", response.text)
            raise

        response_json = response.json()
        if "data" in response_json:
            # Since X.40.0 list responses are encapsulated in "data" with pagination parameters
            return response_json["data"]

        return response_json

    def find_database(self, name: str) -> Optional[Mapping]:
        """Finds database by name attribute or returns none."""
        for api_database in list(self._api("get", "/api/database")):
            if api_database["name"].upper() == name.upper():
                return api_database
        return None

    def sync_database_schema(self, uid: str):
        """Triggers schema sync on a database."""
        self._api("post", f"/api/database/{uid}/sync_schema")

    def get_database_metadata(self, uid: str) -> Mapping:
        """Retrieves metadata for all tables and fields in a database, including hidden ones."""
        return dict(
            self._api(
                method="get",
                path=f"/api/database/{uid}/metadata",
                params={"include_hidden": True},
            )
        )

    def get_tables(self) -> Sequence[Mapping]:
        """Retrieves all tables for all databases."""
        return list(self._api("get", "/api/table"))

    def get_collections(self, exclude_personal: bool) -> Sequence[Mapping]:
        """Retrieves all collections and optionally filters out personal collections."""
        results = list(
            self._api(
                method="get",
                path="/api/collection",
                params={"exclude-other-user-collections": exclude_personal},
            )
        )
        if exclude_personal:
            results = list(filter(lambda x: not x.get("personal_owner_id"), results))
        return results

    def get_collection_items(
        self,
        uid: str,
        models: Sequence[str],
    ) -> Sequence[Mapping]:
        """Retrieves collection items of specific types (e.g. card, dashboard, collection)."""
        results = list(
            self._api(
                method="get",
                path=f"/api/collection/{uid}/items",
                params={"models": models},
            )
        )
        results = list(filter(lambda x: x["model"] in models, results))
        return results

    def find_card(self, uid: str) -> Optional[Mapping]:
        """Retrieves card (known as question in Metabase UI)."""
        try:
            return dict(self._api("get", f"/api/card/{uid}"))
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                _logger.warning("Card '%s' not found", uid)
                return None
            raise

    def format_card_url(self, uid: str) -> str:
        """Formats URL link to a card (known as question in Metabase UI)."""
        return f"{self.url}/card/{uid}"

    def find_dashboard(self, uid: str) -> Optional[Mapping]:
        """Retrieves dashboard."""
        try:
            return dict(self._api("get", f"/api/dashboard/{uid}"))
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                _logger.warning("Dashboard '%s' not found", uid)
                return None
            raise

    def format_dashboard_url(self, uid: str) -> str:
        """Formats URL link to a dashboard."""
        return f"{self.url}/dashboard/{uid}"

    def find_user(self, uid: str) -> Optional[Mapping]:
        """Finds user by ID or returns none."""
        try:
            return dict(self._api("get", f"/api/user/{uid}"))
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                _logger.warning("User '%s' not found", uid)
                return None
            raise

    def update_table(self, uid: str, body: Mapping) -> Mapping:
        """Posts update to an existing table."""
        return dict(self._api("put", f"/api/table/{uid}", json=body))

    def update_table_field_order(self, uid: str, body: Sequence) -> Sequence:
        """Posts update to field order of an existing table."""
        return list(self._api("put", f"/api/table/{uid}/fields/order", json=body))

    def update_field(self, uid: str, body: Mapping) -> Mapping:
        """Posts an update to an existing table field."""
        return dict(self._api("put", f"/api/field/{uid}", json=body))



import unittest
from operator import itemgetter
from pathlib import Path

import yaml

from ._mocks import FIXTURES_PATH, TMP_PATH, MockDbtMetabase


class TestExposures(unittest.TestCase):
    def setUp(self):
        self.c = MockDbtMetabase()
        TMP_PATH.mkdir(exist_ok=True)

    def _assert_exposures(self, expected_path: Path, actual_path: Path):
        with open(expected_path, encoding="utf-8") as f:
            expected = yaml.safe_load(f)
        with open(actual_path, encoding="utf-8") as f:
            actual = yaml.safe_load(f)

        self.assertEqual(
            sorted(expected["exposures"], key=itemgetter("name")),
            actual["exposures"],
        )

    def test_exposures(self):
        fixtures_path = FIXTURES_PATH / "exposure" / "default"
        output_path = TMP_PATH / "exposure" / "default"
        self.c.extract_exposures(
            output_path=str(output_path),
            output_grouping=None,
        )

        self._assert_exposures(
            fixtures_path / "exposures.yml",
            output_path / "exposures.yml",
        )

    def test_exposures_collection_grouping(self):
        fixtures_path = FIXTURES_PATH / "exposure" / "collection"
        output_path = TMP_PATH / "exposure" / "collection"
        self.c.extract_exposures(
            output_path=str(output_path),
            output_grouping="collection",
        )

        self._assert_exposures(
            fixtures_path / "a_look_at_your_customers_table.yml",
            output_path / "a_look_at_your_customers_table.yml",
        )
        self._assert_exposures(
            fixtures_path / "our_analytics.yml",
            output_path / "our_analytics.yml",
        )

    def test_exposures_grouping_type(self):
        fixtures_path = FIXTURES_PATH / "exposure" / "type"
        output_path = TMP_PATH / "exposure" / "type"
        self.c.extract_exposures(
            output_path=str(output_path),
            output_grouping="type",
        )

        for i in [*range(1, 18), 24]:
            self._assert_exposures(
                fixtures_path / "card" / f"{i}.yml",
                output_path / "card" / f"{i}.yml",
            )

        for i in range(1, 2):
            self._assert_exposures(
                fixtures_path / "dashboard" / f"{i}.yml",
                output_path / "dashboard" / f"{i}.yml",
            )

    def test_exposures_aliased_ref(self):
        for model in self.c.manifest.read_models():
            if not model.name.startswith("stg_"):
                model.alias = f"{model.name}_alias"

        aliases = [m.alias for m in self.c.manifest.read_models()]
        self.assertIn("orders_alias", aliases)
        self.assertIn("customers_alias", aliases)

        fixtures_path = FIXTURES_PATH / "exposure" / "default"
        output_path = TMP_PATH / "exposure" / "aliased"
        self.c.extract_exposures(
            output_path=str(output_path),
            output_grouping=None,
        )

        self._assert_exposures(
            fixtures_path / "exposures.yml",
            output_path / "exposures.yml",
        )



import unittest

from dbtmetabase.format import Filter, NullValue, dump_yaml, safe_description, safe_name

from ._mocks import FIXTURES_PATH, TMP_PATH


class TestFormat(unittest.TestCase):
    def test_filter(self):
        self.assertTrue(
            Filter(
                include=("alpHa", "bRavo"),
            ).match("Alpha")
        )
        self.assertTrue(Filter().match("Alpha"))
        self.assertTrue(Filter().match(""))
        self.assertFalse(
            Filter(
                include=("alpHa", "bRavo"),
                exclude=("alpha",),
            ).match("Alpha")
        )
        self.assertFalse(
            Filter(
                exclude=("alpha",),
            ).match("Alpha")
        )
        self.assertTrue(Filter(include="alpha").match("Alpha"))
        self.assertFalse(Filter(exclude="alpha").match("Alpha"))

    def test_filter_wildcard(self):
        self.assertTrue(Filter(include="stg_*").match("stg_orders"))
        self.assertTrue(Filter(include="STG_*").match("stg_ORDERS"))
        self.assertFalse(Filter(include="stg_*").match("orders"))
        self.assertTrue(Filter(include="order?").match("orders"))
        self.assertFalse(Filter(include="order?").match("ordersz"))
        self.assertTrue(Filter(include="*orders", exclude="stg_*").match("_orders"))
        self.assertFalse(Filter(include="*orders", exclude="stg_*").match("stg_orders"))

    def test_null_value(self):
        self.assertIsNotNone(NullValue)
        self.assertFalse(NullValue)
        self.assertIs(NullValue, NullValue)

    def test_safe_name(self):
        self.assertEqual(
            "somebody_s_2_collections_",
            safe_name("Somebody's 2 collections!"),
        )
        self.assertEqual(
            "somebody_s_2_collections_",
            safe_name("somebody_s_2_collections_"),
        )
        self.assertEqual("", safe_name(""))

    def test_safe_description(self):
        self.assertEqual(
            "Depends on\n\nQuestion ( #2 )!",
            safe_description("Depends on\n\nQuestion {{ #2 }}!"),
        )
        self.assertEqual(
            "Depends on\n\nQuestion ( #2 )!",
            safe_description("Depends on\n\nQuestion ( #2 )!"),
        )
        self.assertEqual(
            "Depends on\n\nQuestion { #2 }!",
            safe_description("Depends on\n\nQuestion { #2 }!"),
        )
        self.assertEqual(
            "(start_date) - cast((rolling_days))",
            safe_description("{{start_date}} - cast({{rolling_days}})"),
        )

    def test_dump_yaml(self):
        fixture_path = FIXTURES_PATH / "test_dump_yaml.yml"
        output_path = TMP_PATH / "test_dump_yaml.yml"
        with open(output_path, "w", encoding="utf-8") as f:
            dump_yaml(
                data={
                    "root": {
                        "attr1": "val1\nend",
                        "attr2": ["val2", "val3"],
                    },
                },
                stream=f,
            )
        with open(output_path, "r", encoding="utf-8") as f:
            actual = f.read()
        with open(fixture_path, "r", encoding="utf-8") as f:
            expected = f.read()
        self.assertEqual(expected, actual)



import unittest
from operator import attrgetter
from typing import Optional, Sequence

from dbtmetabase.manifest import Column, Group, Manifest, Model

from ._mocks import FIXTURES_PATH


class TestManifest(unittest.TestCase):
    def test_v11_disabled(self):
        models = Manifest(FIXTURES_PATH / "manifest-v11-disabled.json").read_models()

        orders_mod = self._find_model(models, "orders")
        self.assertIsNone(orders_mod)

        customer_id_col = self._find_column(models, "customers", "customer_id")
        assert customer_id_col
        self.assertIsNone(customer_id_col.fk_target_table)
        self.assertIsNone(customer_id_col.fk_target_field)

    def test_v11(self):
        models = Manifest(FIXTURES_PATH / "manifest-v11.json").read_models()
        self._assertModelsEqual(
            models,
            [
                Model(
                    database="dbtmetabase",
                    schema="public",
                    group=Group.nodes,
                    name="customers",
                    alias="customers",
                    description="This table has basic information about a customer, as well as some derived facts based on a customer's orders",
                    display_name="clients",
                    unique_id="model.sandbox.customers",
                    columns=[
                        Column(
                            name="customer_id",
                            description="This is a unique identifier for a customer",
                        ),
                        Column(
                            name="first_name",
                            description="Customer's first name. PII.",
                        ),
                        Column(
                            name="last_name",
                            description="Customer's last name. PII.",
                        ),
                        Column(
                            name="first_order",
                            description="Date (UTC) of a customer's first order",
                        ),
                        Column(
                            name="most_recent_order",
                            description="Date (UTC) of a customer's most recent order",
                        ),
                        Column(
                            name="number_of_orders",
                            description="Count of the number of orders a customer has placed",
                            display_name="order_count",
                        ),
                        Column(
                            name="customer_lifetime_value",
                            description="Total value (AUD) of a customer's orders",
                        ),
                    ],
                ),
                Model(
                    database="dbtmetabase",
                    schema="public",
                    group=Group.nodes,
                    name="orders",
                    alias="orders",
                    description="This table has basic information about orders, as well as some derived facts based on payments",
                    points_of_interest="Basic information only",
                    caveats="Some facts are derived from payments",
                    unique_id="model.sandbox.orders",
                    columns=[
                        Column(
                            name="order_id",
                            description="This is a unique identifier for an order",
                            semantic_type="type/PK",
                        ),
                        Column(
                            name="customer_id",
                            description="Foreign key to the customers table",
                            semantic_type="type/FK",
                            fk_target_table="public.customers",
                            fk_target_field="customer_id",
                        ),
                        Column(
                            name="order_date",
                            description="Date (UTC) that the order was placed",
                        ),
                        Column(
                            name="status",
                            description="",
                        ),
                        Column(
                            name="amount",
                            description="Total amount (AUD) of the order",
                        ),
                        Column(
                            name="credit_card_amount",
                            description="Amount of the order (AUD) paid for by credit card",
                        ),
                        Column(
                            name="coupon_amount",
                            description="Amount of the order (AUD) paid for by coupon",
                        ),
                        Column(
                            name="bank_transfer_amount",
                            description="Amount of the order (AUD) paid for by bank transfer",
                        ),
                        Column(
                            name="gift_card_amount",
                            description="Amount of the order (AUD) paid for by gift card",
                        ),
                    ],
                ),
                Model(
                    database="dbtmetabase",
                    schema="public",
                    group=Group.nodes,
                    name="stg_customers",
                    alias="stg_customers",
                    description="",
                    unique_id="model.sandbox.stg_customers",
                    columns=[
                        Column(
                            name="customer_id",
                            description="",
                        ),
                        Column(
                            name="first_name",
                            description="",
                        ),
                        Column(
                            name="last_name",
                            description="",
                        ),
                    ],
                ),
                Model(
                    database="dbtmetabase",
                    schema="public",
                    group=Group.nodes,
                    name="stg_payments",
                    alias="stg_payments",
                    description="",
                    unique_id="model.sandbox.stg_payments",
                    columns=[
                        Column(
                            name="payment_id",
                            description="",
                        ),
                        Column(
                            name="payment_method",
                            description="",
                        ),
                        Column(
                            name="order_id",
                            description="",
                        ),
                        Column(
                            name="amount",
                            description="",
                        ),
                    ],
                ),
                Model(
                    database="dbtmetabase",
                    schema="public",
                    group=Group.nodes,
                    name="stg_orders",
                    alias="stg_orders",
                    description="",
                    unique_id="model.sandbox.stg_orders",
                    columns=[
                        Column(
                            name="order_id",
                            description="",
                        ),
                        Column(
                            name="status",
                            description="",
                        ),
                        Column(
                            name="order_date",
                            description="",
                        ),
                        Column(
                            name="customer_id",
                            description="",
                        ),
                    ],
                ),
            ],
        )

    def test_v2(self):
        models = Manifest(FIXTURES_PATH / "manifest-v2.json").read_models()
        self._assertModelsEqual(
            models,
            [
                Model(
                    database="test",
                    schema="public",
                    group=Group.nodes,
                    name="orders",
                    alias="orders",
                    description="This table has basic information about orders, as well as some derived facts based on payments",
                    unique_id="model.jaffle_shop.orders",
                    columns=[
                        Column(
                            name="order_id",
                            description="This is a unique identifier for an order",
                        ),
                        Column(
                            name="customer_id",
                            description="Foreign key to the customers table",
                            semantic_type="type/FK",
                            fk_target_table="public.customers",
                            fk_target_field="customer_id",
                        ),
                        Column(
                            name="order_date",
                            description="Date (UTC) that the order was placed",
                        ),
                        Column(
                            name="status",
                            description="Orders can be one of the following statuses:\n\n| status         | description                                                                                                            |\n|----------------|------------------------------------------------------------------------------------------------------------------------|\n| placed         | The order has been placed but has not yet left the warehouse                                                           |\n| shipped        | The order has ben shipped to the customer and is currently in transit                                                  |\n| completed      | The order has been received by the customer                                                                            |\n| return_pending | The customer has indicated that they would like to return the order, but it has not yet been received at the warehouse |\n| returned       | The order has been returned by the customer and received at the warehouse                                              |",
                        ),
                        Column(
                            name="amount",
                            description="Total amount (AUD) of the order",
                        ),
                        Column(
                            name="credit_card_amount",
                            description="Amount of the order (AUD) paid for by credit card",
                        ),
                        Column(
                            name="coupon_amount",
                            description="Amount of the order (AUD) paid for by coupon",
                        ),
                        Column(
                            name="bank_transfer_amount",
                            description="Amount of the order (AUD) paid for by bank transfer",
                        ),
                        Column(
                            name="gift_card_amount",
                            description="Amount of the order (AUD) paid for by gift card",
                        ),
                    ],
                ),
                Model(
                    database="test",
                    schema="public",
                    group=Group.nodes,
                    name="customers",
                    alias="customers",
                    description="This table has basic information about a customer, as well as some derived facts based on a customer's orders",
                    unique_id="model.jaffle_shop.customers",
                    columns=[
                        Column(
                            name="customer_id",
                            description="This is a unique identifier for a customer",
                            semantic_type=None,  # This is a PK field, should not be detected as FK
                        ),
                        Column(
                            name="first_name",
                            description="Customer's first name. PII.",
                        ),
                        Column(
                            name="last_name",
                            description="Customer's last name. PII.",
                        ),
                        Column(
                            name="first_order",
                            description="Date (UTC) of a customer's first order",
                        ),
                        Column(
                            name="most_recent_order",
                            description="Date (UTC) of a customer's most recent order",
                        ),
                        Column(
                            name="number_of_orders",
                            description="Count of the number of orders a customer has placed",
                        ),
                        Column(
                            name="customer_lifetime_value",
                            description="Total value (AUD) of a customer's orders",
                        ),
                    ],
                ),
                Model(
                    database="test",
                    schema="public",
                    group=Group.nodes,
                    name="stg_orders",
                    alias="stg_orders",
                    description="",
                    unique_id="model.jaffle_shop.stg_orders",
                    columns=[
                        Column(
                            name="order_id",
                            description="",
                        ),
                        Column(
                            name="status",
                            description="",
                        ),
                    ],
                ),
                Model(
                    database="test",
                    schema="public",
                    group=Group.nodes,
                    name="stg_payments",
                    alias="stg_payments",
                    description="",
                    unique_id="model.jaffle_shop.stg_payments",
                    columns=[
                        Column(
                            name="payment_id",
                            description="",
                        ),
                        Column(
                            name="payment_method",
                            description="",
                        ),
                    ],
                ),
                Model(
                    database="test",
                    schema="public",
                    group=Group.nodes,
                    name="stg_customers",
                    alias="stg_customers",
                    description="",
                    unique_id="model.jaffle_shop.stg_customers",
                    tags=[],
                    columns=[
                        Column(
                            name="customer_id",
                            description="",
                        )
                    ],
                ),
            ],
        )

    def _assertModelsEqual(
        self,
        first: Sequence[Model],
        second: Sequence[Model],
    ):
        self.assertEqual(len(first), len(second), "mismatched model count")

        first = sorted(first, key=attrgetter("name"))
        second = sorted(second, key=attrgetter("name"))

        for i, first_model in enumerate(first):
            second_model = second[i]
            self.assertEqual(first_model.name, second_model.name, "wrong model")
            self.assertEqual(
                len(first_model.columns),
                len(second_model.columns),
                f"mismatched column count in {first_model.name}",
            )
            for j, first_column in enumerate(first_model.columns):
                second_column = second_model.columns[j]
                self.assertEqual(
                    first_column.name,
                    second_column.name,
                    f"wrong column in model {first_model.name}",
                )
                self.assertEqual(
                    first_column,
                    second_column,
                    f"mismatched column {first_model.name}.{first_column.name}",
                )
            self.assertEqual(
                first_model,
                second_model,
                f"mismatched model {first_model.name}",
            )

        self.assertEqual(first, second)

    @staticmethod
    def _find_model(models: Sequence[Model], model_name: str) -> Optional[Model]:
        filtered = [m for m in models if m.name == model_name]
        if filtered:
            return filtered[0]
        return None

    @staticmethod
    def _find_column(
        models: Sequence[Model],
        model_name: str,
        column_name: str,
    ) -> Optional[Column]:
        model = TestManifest._find_model(models=models, model_name=model_name)
        if model:
            filtered = [c for c in model.columns if c.name == column_name]
            if filtered:
                return filtered[0]
        return None



#!/usr/bin/env python3

import sys

from setuptools import find_packages, setup

if sys.version_info < (3, 8):
    raise ValueError("Requires Python 3.8+")


def requires_from_file(filename: str) -> list:
    with open(filename, "r", encoding="utf-8") as f:
        return [x.strip() for x in f if x.strip()]


with open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="dbt-metabase",
    description="dbt + Metabase integration.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Mike Gouline",
    url="https://github.com/gouline/dbt-metabase",
    license="MIT License",
    entry_points={
        "console_scripts": ["dbt-metabase = dbtmetabase.__main__:cli"],
    },
    packages=find_packages(exclude=["tests", "sandbox"]),
    test_suite="tests",
    install_requires=requires_from_file("requirements.txt"),
    tests_require=requires_from_file("requirements-test.txt"),
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import unittest

from ._mocks import MockDbtMetabase


class TestModels(unittest.TestCase):
    def setUp(self):
        # pylint: disable=protected-access
        self.c = MockDbtMetabase()
        self.c._ModelsMixin__SYNC_PERIOD = 1  # type: ignore

    def test_export(self):
        self.c.export_models(
            metabase_database="unit_testing",
            skip_sources=True,
            sync_timeout=0,
            order_fields=True,
        )

    def test_build_lookups(self):
        # pylint: disable=protected-access,no-member
        expected = {
            "PUBLIC.CUSTOMERS": [
                "CUSTOMER_ID",
                "FIRST_NAME",
                "LAST_NAME",
                "FIRST_ORDER",
                "MOST_RECENT_ORDER",
                "NUMBER_OF_ORDERS",
                "CUSTOMER_LIFETIME_VALUE",
            ],
            "PUBLIC.ORDERS": [
                "ORDER_ID",
                "CUSTOMER_ID",
                "ORDER_DATE",
                "STATUS",
                "CREDIT_CARD_AMOUNT",
                "COUPON_AMOUNT",
                "BANK_TRANSFER_AMOUNT",
                "GIFT_CARD_AMOUNT",
                "AMOUNT",
            ],
            "PUBLIC.RAW_CUSTOMERS": ["ID", "FIRST_NAME", "LAST_NAME"],
            "PUBLIC.RAW_ORDERS": ["ID", "USER_ID", "ORDER_DATE", "STATUS"],
            "PUBLIC.RAW_PAYMENTS": ["ID", "ORDER_ID", "PAYMENT_METHOD", "AMOUNT"],
            "PUBLIC.STG_CUSTOMERS": ["CUSTOMER_ID"],
            "PUBLIC.STG_ORDERS": ["ORDER_ID", "STATUS"],
            "PUBLIC.STG_PAYMENTS": ["PAYMENT_ID", "PAYMENT_METHOD"],
        }

        actual_tables = self.c._ModelsMixin__get_tables(database_id="2")  # type: ignore

        self.assertEqual(list(expected.keys()), list(actual_tables.keys()))

        for table, columns in expected.items():
            self.assertEqual(columns, list(actual_tables[table]["fields"].keys()))



from __future__ import annotations

import fnmatch
import logging
import re
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Any, MutableSequence, Optional, Sequence, TextIO

import yaml
from rich.logging import RichHandler


class Filter:
    """Inclusion/exclusion filtering."""

    def __init__(
        self,
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
    ):
        """Inclusion/exclusion filtering.

        Args:
            include (Optional[Sequence[str]], optional): Optional inclusions (i.e. include only these). Defaults to None.
            exclude (Optional[Sequence[str]], optional): Optional exclusion list (i.e. exclude these, even if in inclusion list). Defaults to None.
        """
        self.include = self._norm_arg(include)
        self.exclude = self._norm_arg(exclude)

    def match(self, item: str) -> bool:
        item = self._norm_item(item)

        for exclude in self.exclude:
            if fnmatch.fnmatch(item, exclude):
                return False

        if self.include:
            for include in self.include:
                if fnmatch.fnmatch(item, include):
                    return True
            return False

        return True

    @staticmethod
    def _norm_arg(arg: Optional[Sequence[str]]) -> Sequence[str]:
        if isinstance(arg, str):
            arg = [arg]
        return [Filter._norm_item(x) for x in arg or []]

    @staticmethod
    def _norm_item(x: str) -> str:
        return x.upper()


class _YAMLDumper(yaml.Dumper):
    """Custom YAML dumper for uniform formatting."""

    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, indentless=False)


class _NullValue(str):
    """Explicitly null field value."""

    def __eq__(self, other: object) -> bool:
        return other is None


NullValue = _NullValue()


def dump_yaml(data: Any, stream: TextIO):
    """Uniform way to dump object to YAML file.

    Args:
        data (Any): Payload.
        stream (TextIO): Text file handle.
    """
    yaml.dump(
        data,
        stream,
        Dumper=_YAMLDumper,
        default_flow_style=False,
        allow_unicode=True,
        sort_keys=False,
    )


def setup_logging(level: int, path: Optional[Path] = None):
    """Basic logger configuration for the CLI.

    Args:
        level (int): Logging level. Defaults to logging.INFO.
        path (Path): Path to file logs.
    """

    handlers: MutableSequence[logging.Handler] = []

    if path:
        path.parent.mkdir(parents=True, exist_ok=True)
        file_handler = RotatingFileHandler(
            filename=path,
            maxBytes=int(1e6),
            backupCount=3,
        )
        file_handler.setFormatter(
            logging.Formatter("%(asctime)s — %(name)s — %(levelname)s — %(message)s")
        )
        file_handler.setLevel(logging.WARNING)
        handlers.append(file_handler)

    handlers.append(
        RichHandler(
            level=level,
            rich_tracebacks=True,
            markup=True,
            show_time=False,
        )
    )

    logging.basicConfig(
        level=level,
        format="%(asctime)s — %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S %z",
        handlers=handlers,
        force=True,
    )


def safe_name(text: Optional[str]) -> str:
    """Sanitizes a human-readable "friendly" name to a safe string.

    For example, "Joe's Collection" becomes "joe_s_collection".

    Args:
        text (Optional[str]): Unsafe text with non-underscore symbols and spaces.

    Returns:
        str: Sanitized lowercase string with underscores.
    """
    return re.sub(r"[^\w]", "_", text or "").lower()


def safe_description(text: Optional[str]) -> str:
    """Sanitizes a human-readable long text, such as description.

    Args:
        text (Optional[str]): Unsafe long text with Jinja syntax.

    Returns:
        str: Sanitized string with escaped Jinja syntax.
    """
    return re.sub(r"{{(.*?)}}", r"(\1)", text or "")
